# -*- coding: UTF-8 -*-
#!/usr/bin/env python3

import wx, wx.adv, sys, os
import wx.lib.agw.flatnotebook as fnb
import cv2, numpy as np
import random, math
import re
import time


from argparse import ArgumentParser

import teraFAST.gui
import teraFAST.ref as ref
import teraFAST.processor as  processor
import teraFAST.worker as worker

import threading

import zipfile

from teraFAST.gui.bcgpanel import *
from teraFAST.gui.optionspanel import *

# pixel proportions
pxRATIO = 192./192.
pxSCALE  = 1

# Button definitions

DEFAULT_CHUNK = 256
DEFAULT_ROWS = 512
DEFAULT_RATE = 3000

INITIAL_IMG_SIZE = 256  #vertical size, pixels
#ASPECT_RATIO = 4
ASPECT_RATIO = 2

MIN_VIDEO_OUT = 16

IMG_PREFIX = 'scr_'

FAVICON_NAME = 'tera.ico'

UpdateEventType = wx.NewEventType()
EVT_UPDATE = wx.PyEventBinder(UpdateEventType, 1)

version = ""
commandline_options = None

class UpdateEvent(wx.PyCommandEvent):
    """Simple event to carry arbitrary result data."""
    def __init__(self, data=None):
        wx.PyCommandEvent.__init__(self, UpdateEventType)
        self.data = data
        

class ImageDisplay(wx.Panel):
    def __init__(self, parent, id, pos=(0,0), size=(400,400), style=0, name="",imgSize=(16,16)):

        self.parent = parent
        wx.Panel.__init__(self, parent, id,  pos, size, style, name)
        self.SetBackgroundStyle(wx.BG_STYLE_CUSTOM)
        self.DCsize = tuple(self.GetClientSize())

        self.size = imgSize

        try:
            self.bitmap = wx.Bitmap(*self.DCsize)
        except:
            self.bitmap = wx.EmptyBitmap(*self.DCsize)
        self.img = np.zeros(self.DCsize+(3,),dtype=np.uint8)
        self.scaledImg = np.empty(self.DCsize+(3,),dtype=np.uint8)
        self.grid = self.createGrid()
            
        self.startPos = self.endPos = None
        
        self.drawGrid = False
        
        try:
            self.buffer = wx.Bitmap(*self.DCsize)
        except:
            self.buffer = wx.EmptyBitmap(*self.DCsize)
        
        
        self.Bind(wx.EVT_SIZE, self.OnSize)
        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.Update(self.img)
        self.SetCursor(wx.Cursor(wx.CURSOR_ARROW))

        self.Bind(wx.EVT_LEFT_DOWN, self.OnLeftDown) 
        self.Bind(wx.EVT_LEFT_UP, self.OnLeftUp) 
        self.Bind(wx.EVT_RIGHT_DOWN, self.OnRightDown)


        self.SetMinSize(size)

    def createGrid(self):
        try:
            color = cv2.cv.RGB(128,128,128)
        except:
            color = (128,128,128)
        grid = np.zeros_like(self.scaledImg,dtype=np.uint8)
        n = (self.size[1]//4)
        dx = float(grid.shape[0])/n
        for i in range(n+1):
            xi = int(i*dx)
            cv2.line(grid,(0,xi),(grid.shape[1],xi),color)
        n = (self.size[0]//8)
        dx = float(grid.shape[1])/n
        for i in range(n+1):
            xi = int(i*dx)
            cv2.line(grid,(xi,0),(xi,grid.shape[0]),color)
        return grid 

    def Draw(self, dc=None):
        """ Draw image """
        if dc is None:
            dc = wx.BufferedDC(wx.ClientDC(self),self.buffer)
        dc.DrawBitmap(self.bitmap,0,0)
        if self.startPos is not None:
            dc.SetPen(wx.Pen("white", 1))
            dc.SetBrush(wx.TRANSPARENT_BRUSH)
            try:
                dc.DrawRectangle( self.RectAround(self.startPos, self.endPos) )
            except:
                dc.DrawRectangleRect( self.RectAround(self.startPos, self.endPos) )
                

    
    def OnPaint(self, event=None):
        """ Process paint event """
        dc = wx.AutoBufferedPaintDC(self)
        self.Draw(dc)

    def OnSize(self, event=None):
        """ Process resize event """
        self.DCsize = tuple(self.GetClientSize())
        self.scaledImg = np.empty(self.DCsize+(3,),dtype=np.uint8)
        self.grid = self.createGrid()
        try:
            self.buffer = wx.Bitmap(*self.DCsize)
        except:
            self.buffer = wx.EmptyBitmap(*self.DCsize)
        self.Update(self.img.transpose([1,0,2]))
        
    
    def Update(self, img):
        self.img = img.transpose([1,0,2]).copy()
##        self.img = img
        self.scaledImg=cv2.resize(self.img,self.DCsize,interpolation=cv2.INTER_NEAREST)
        if self.drawGrid:
            self.scaledImg = cv2.add(self.scaledImg,self.grid)
        try:
            self.bitmap = wx.Bitmap(wx.Image(*(self.DCsize+(bytearray(self.scaledImg.tobytes()),))))
        except:            
            self.bitmap = wx.BitmapFromImage(wx.ImageFromData(*(self.DCsize+(bytearray(self.scaledImg.tostring()),))))
        self.Draw(None)

    def OnLeftDown(self, evt): 
         # Capture the mouse and save the starting posiiton for the 
         # rubber-band 
        self.CaptureMouse() 
        self.startPos = self.endPos = self.PointToImgCoords(evt.GetPosition())
        self.Bind(wx.EVT_MOTION, self.OnMouseMove)
        evt.Skip()


    def OnMouseMove(self, evt): 
        if evt.Dragging() and evt.LeftIsDown():
            p = self.PointToImgCoords(evt.GetPosition())
            if self.endPos != p:
                self.endPos = p
                self.Draw(None)
               
#                self.parent.OnUpdate(UpdateEvent(False))
                evt.Skip()
                wx.PostEvent(self, UpdateEvent(False))
        else:
            evt.Skip()

    def OnLeftUp(self, evt): 
        if self.HasCapture(): 
             self.ReleaseMouse()        
        self.Unbind(wx.EVT_MOTION) 
        self.Draw(None)
        evt.Skip()
        wx.PostEvent(self, UpdateEvent(True)) 

    def OnRightDown(self, evt): 
         # restores the former content to the window. 
         self.startPos = self.endPos = None
         self.Draw(None)
         evt.Skip()
         wx.PostEvent(self, UpdateEvent(True))

    def GetSelection(self):
        if self.startPos is None:
            return (None, None)
        x1,y1 = self.startPos
        x2, y2 = self.endPos
        return (max(0,min(x1,x2)), max(0,min(y1,y2)) ), ( min(self.size[0]-1,max(x1,x2)) ,  min(self.size[1]-1,max(y1,y2)))
    
    def PointToImgCoords(self, point):
        return (int((point[0]*self.size[0]//self.DCsize[0])),int((point[1]*self.size[1]//self.DCsize[1])))

    def RectAround(self,p1,p2):
        x1 = int(max(0,min(p1[0],p2[0]))*self.DCsize[0]/self.size[0])
        x2 = int(min(self.size[0],max(p1[0],p2[0])+1)*self.DCsize[0]/self.size[0])
        
        y1 = int(max(0,min(p1[1],p2[1]))*self.DCsize[1]/self.size[1])
        y2 = int(min(self.size[1],max(p1[1],p2[1])+1)*self.DCsize[1]/self.size[1])
        
        return wx.Rect(x1,y1,x2-x1,y2-y1)
        
#End of class ImageDisplay(wx.Panel)

       
# GUI Frame class that spins off the worker thread
class MainFrame(wx.Frame):
    """Class MainFrame."""
    def __init__(self, parent, id, style=wx.DEFAULT_FRAME_STYLE, sn=''):
        """Create the MainFrame."""

       
        self.flags = 0
         
        wx.Frame.__init__(self, parent, id, 'TeraFAST View', style=style, size=(100,100))

        self.status = self.CreateStatusBar(number=5,style=wx.SB_RAISED)

        favicon = wx.Icon(os.path.join(os.path.dirname(os.path.abspath(teraFAST.gui.__file__)),FAVICON_NAME), wx.BITMAP_TYPE_ICO)
        self.SetIcon(favicon)

        self.counter  = 0

        self.fps  = 0

        self.interface = None
        self.videoWriter = None
        self.selection = (None,None) #
        self.values = {"mean":0.,"min":0.,"max":0.} #mean, min, max

        self.savePath = None

        self._stopme = False
# Create interface object

        try:
            defaults = __import__("config",globals(),locals(),['config']).config
        except:
            defaults = None
            
        
        try:
            config = os.path.join(os.path.dirname(__file__),defaults.configPath)
        except:
            config = None

        try:
            self.dev = defaults.developmentMode
        except:
            self.dev = False

        
        try:
            self.processor = processor.processor(rows=DEFAULT_CHUNK, threaded = True, defaults=defaults,config=config, sn=sn)
        except RuntimeError as e:
            self.OnError(e)
            return
        self.sn = self.processor.GetID()
        
        self.status.SetStatusText(self.processor.interface.info(),0)

        try:
            self.status.SetStatusText(self.sn,4)
        except:
            pass

#        rows = DEFAULT_ROWS if self.processor.X_SIZE*ASPECT_RATIO > DEFAULT_ROWS else self.processor.X_SIZE*ASPECT_RATIO
        rows = int(self.processor.X_SIZE*ASPECT_RATIO)
        pxSCALE = INITIAL_IMG_SIZE/self.processor.X_SIZE
        size = (int(round(rows*pxSCALE)), int(round(self.processor.X_SIZE*pxSCALE*pxRATIO)))
        imgSize = (rows, self.processor.X_SIZE)
        self.imgSize = imgSize  ###

        if DEFAULT_CHUNK > rows:
            self.processor.SetFrameLength(rows)
        
        self.image = ImageDisplay(self, -1, size = size, style = wx.DOUBLE_BORDER|wx.RESIZE_BORDER, imgSize = imgSize)
        self.img = np.zeros(imgSize+(3,),dtype=np.uint8)
        self.data = np.zeros(imgSize,dtype=np.float32)

        
        
       
        self.worker = worker.Worker(imgSize)



                

        self.nb = fnb.FlatNotebook(self, -1, agwStyle = fnb.FNB_NO_X_BUTTON | fnb.FNB_NO_NAV_BUTTONS )

    #   

        self.SetBackgroundColour('#f0f0f0')

        self.saveimg = wx.Button(self,-1, 'Save image...')
        self.saveimg.Disable()

        self.writevideo = wx.Button(self,-1, 'Save video...')
        self.writevideo.Disable()
        
        self.startstop = wx.Button(self, wx.ID_ANY, 'Start')

        
        self.panelBCG = BCGPanel(self.nb, -1, main = self, processor = self.processor)
        self.panelOptions = OptionsPanel(self.nb, -1,main = self, processor = self.processor)
		

        self.nb.AddPage(self.panelBCG, "B/C/G")
        self.nb.AddPage(self.panelOptions, "Options")

        

#  Menu
        self.menubar = wx.MenuBar()
        self.fileMenu = wx.Menu()
        mi = wx.MenuItem(self.fileMenu,wx.ID_SAVE, '&Save Image\tCtrl+S', 'Save current image')
        
        self.Bind(wx.EVT_MENU, self.OnSaveImg, id=mi.GetId())

        mi = wx.MenuItem(self.fileMenu,-1, '&Save All\tF5', 'Save all formats')
        try:
            self.fileMenu.Append(mi)
        except:
            self.fileMenu.AppendItem(mi)
        self.Bind(wx.EVT_MENU, self.OnSaveAll, id=mi.GetId())


        mi = wx.MenuItem(self.fileMenu,-1, '&Run\tCtrl+R', 'Start')
        try:
            self.fileMenu.Append(mi)
        except:
            self.fileMenu.AppendItem(mi)
        self.Bind(wx.EVT_MENU, self.OnStart, id=mi.GetId())

        mi = wx.MenuItem(self.fileMenu,-1, '&Pause\tCtrl+P', 'Stop')
        try:
            self.fileMenu.Append(mi)
        except:
            self.fileMenu.AppendItem(mi)
        self.Bind(wx.EVT_MENU, self.OnStop, id=mi.GetId())

        self.fileMenu.AppendSeparator()
        
        mi = wx.MenuItem(self.fileMenu,wx.ID_EXIT, '&Quit\tCtrl+Q', 'Quit application')
        try:
            self.fileMenu.Append(mi)
        except:
            self.fileMenu.AppendItem(mi)
        self.Bind(wx.EVT_MENU, self.OnQuit, id=mi.GetId())
        
        self.menubar.Append(self.fileMenu, '&File')

        
        self.helpMenu = wx.Menu()
        mi = wx.MenuItem(self.helpMenu,wx.ID_ABOUT, '&About...\tF10', 'About the app')
        try:
            self.helpMenu.Append(mi)
        except:
            self.helpMenu.AppendItem(mi)
        self.Bind(wx.EVT_MENU, self.OnAboutBox, id=mi.GetId())
        
        self.menubar.Append(self.helpMenu, '&Help')
        
        
        self.SetMenuBar(self.menubar)
        
#        self.Bind(wx.EVT_MENU, self.OnQuit, id=wx.ID_EXIT)


        self.statsBox = wx.StaticBox(self, label = "Selection")
        self.statsData = wx.StaticText(self, label="0.000")
#        self.statsData.SetLabel("Min:0.234\tMin:0.234\tMin:0.234x")
        font = self.statsData.GetFont()
#        font.SetWeight(wx.FONTWEIGHT_BOLD)
        font.SetPointSize(int(font.GetPointSize()*1.3))
        self.statsData.SetFont(font)
        self.statsData.SetMinSize(self.statsData.GetTextExtent("Min: 0.000"+4*" "+"Max: 0.000"+4*" "+"Average: 0.000e-00"))
#        self.statsBox.Show(False)
#        self.statsData.Show(False)

#  Assemble
              
        hbox = wx.BoxSizer(wx.HORIZONTAL)
        self.statsSizer = wx.StaticBoxSizer(self.statsBox,wx.HORIZONTAL)
        self.statsSizer.Add(self.statsData,proportion=0,flag=wx.ALIGN_LEFT|wx.RIGHT|wx.LEFT,border=5)

        hbox.Add(self.statsSizer,proportion=0,flag=wx.ALIGN_LEFT|wx.EXPAND)
        
        hbox.Add((1,0),proportion=1)
       
        hbox.Add(self.saveimg, proportion = 0,flag=wx.EXPAND)
        hbox.Add(self.writevideo, proportion = 0,flag=wx.EXPAND)
        
        hbox.Add((1,0),proportion=1)
        
        hbox.Add(self.startstop, proportion = 0,flag=wx.EXPAND)
        
        sizer3 = wx.BoxSizer(wx.VERTICAL)
        sizer3.Add( self.image,proportion=0, flag=wx.EXPAND|wx.SHAPED|wx.ALIGN_TOP|wx.ALL^wx.RIGHT,border=10)        
        sizer3.Add(hbox,proportion=0, flag=wx.EXPAND|wx.ALIGN_LEFT|wx.ALL,border=10)
        

        
        sizer1 = wx.BoxSizer( wx.HORIZONTAL)        
        
        sizer1.Add(sizer3,proportion=4, flag=wx.EXPAND|wx.ALIGN_TOP|wx.ALL,border=0)
        sizer1.Add( self.nb, proportion=1, flag=wx.EXPAND|wx.ALIGN_TOP|wx.LEFT|wx.RIGHT,border=10)
        
        
    
        self.main_sizer  = sizer1
        self.SetSizerAndFit(self.main_sizer)
        self.main_sizer.SetSizeHints(self)
        
        self.currentSize = self.GetSize()

        
        self.Bind(wx.EVT_SIZE, self.OnSize)

        self.Bind(wx.EVT_BUTTON, self.OnSaveImg,id=self.saveimg.GetId())
        self.Bind(wx.EVT_BUTTON, self.OnWriteVideo,id=self.writevideo.GetId())
         
        self.Bind(wx.EVT_BUTTON, self.OnStart, id=self.startstop.GetId())
        self.Bind(wx.EVT_CLOSE, self.OnStop)

##        self.Bind(wx.EVT_KEY_DOWN, self.onKeyPress)

        # Set up event handler for any worker thread results
        self.Bind(EVT_UPDATE,self.OnUpdate)

        self.lastTime = 0


##    def onKeyPress(self,event):
##        if event.GetKeyCode()==wx.WXK_F5:
##            self.OnSaveAll()
##        else:
##            event.Skip()

    def OnAboutBox(self, event):
        
        description = """This is a software package supporting TeraFAST devices produced by Terasense, Inc. 
		It's primary purpose is to demonstrate hardware and API options.
"""
        licence = """Copyright (c) 2023 Terasense, Inc.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

        try:
            info = wx.AboutDialogInfo()
        except:
            info = wx.adv.AboutDialogInfo()

        info.SetName('TeraFAST View')
        info.SetVersion(version) 
        info.SetDescription(description)
        info.SetCopyright('(C) 2015-2023 Terasense Development Labs')
        info.SetWebSite('http://www.terasense.com')
        info.SetLicence(licence)
        info.AddDeveloper('Gombo Tsydynzhapov')
        info.AddDeveloper('Oleg Volkov')

        try:
            wx.AboutBox(info)
        except:
            wx.adv.AboutBox(info)

    
    def OnStart(self, event=None):
        """Start Computation."""


        self.counter = 0
        self.fps_list = []
##        self.status.SetFields(["Running...","","",""])
        self.status.SetStatusText("Running...",0)
        self.status.SetStatusText("",1)
        self.status.SetStatusText("",2)
        self.status.SetStatusText("",3)
        
        self.startstop.SetLabel("Stop")
        self.Unbind(wx.EVT_BUTTON, id=self.startstop.GetId())
        self.Bind(wx.EVT_BUTTON, self.OnStop, id=self.startstop.GetId())

#        self.Bind(worker.EVT_HISTOGRAM,self.panelBCG.OnStatData)

        self.saveimg.Enable()
        self.writevideo.Enable()


        self.panelBCG.OnStart()
        self.panelOptions.OnStart()

              
        self.processor.start(self.OnResult,self.OnError)
        self._stopme = False
        self.lastTime =  time.perf_counter()
        


    def OnStop(self, event=None):
        """Stop Computation."""
        if isinstance(event, wx.CloseEvent):
            self.status.SetStatusText("Closing App...",0)
            self.processor.stop(False)
        else:
            if not self._stopme:
                self._stopme = True # Flag the worker thread to stop if running
            if self.counter > 0:
                self.status.SetStatusText("Terminating...",0)            
        if event:
            event.Skip() 
                  
    
    def _OnStop(self, event=None):
##        self.processor.stop()
        
        self.panelBCG.OnStop()
        self.panelOptions.OnStop()
        
##        self.saveimg.Disable()      #disable screenshot
        self.writevideo.Disable()

           
            
        if self.counter > 0:
            self.status.SetStatusText("Stopped",0)

        self.startstop.SetLabel("Start")
        self.Unbind(wx.EVT_BUTTON, id=self.startstop.GetId())
        self.Bind(wx.EVT_BUTTON, self.OnStart, id=self.startstop.GetId())
            
        
        

    def OnQuit(self, event=None):
        self.Close()
            
    def OnSaveImg(self, event=None):
        if  self.img is None:
            return
        im = self.img.copy()
        data = self.worker.source.copy()
        w_data = self.worker.im_src.copy()
        dialog = wx.FileDialog ( None, wildcard = "PNG Image (*.png)|*.png|Image data (*.dat)|*.dat|Original data (*.dat)|*.dat", style = wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT  )
        if self.savePath is not None:
            dialog.SetDirectory(self.savePath)
        if dialog.ShowModal() == wx.ID_OK:
            i = dialog.GetFilterIndex()
            self.savePath = dialog.GetDirectory()
            if i==0:
                try:
                    image = wx.Image(*(im.shape[:2]+(bytearray(im.transpose([1,0,2]).tostring()),)))
                except:                    
                    image = wx.ImageFromData(*(im.shape[:2]+(bytearray(im.transpose([1,0,2]).tostring()),)))
                image.SaveFile(dialog.GetPath(),wx.BITMAP_TYPE_PNG)
            else:
                if i == 1:
##                    x = np.transpose(self.worker.data).copy()
                    x = w_data
                    comment = "Processed image data"
                    with open(dialog.GetPath(),"w") as out:
                        out.write('# '+comment+'\n')
                        out.write('# %dx%d\n' % x.shape)
                        for i in range(x.shape[1]):
                            out.write(",\t".join("%3d" % a for a in x[:,i])+"\n")
                elif i == 2:
##                    x = np.transpose(self.data).copy()
                    x = data
                    comment = "Original image data"
                    with open(dialog.GetPath(),"w") as out:
                        out.write('# '+comment+'\n')
                        out.write('# %dx%d\n' % x.shape)
                        for i in range(x.shape[1]):
                            out.write(",\t".join("%.6f" % a for a in x[:,i])+"\n")

        dialog.Destroy()

            
    def OnSaveAll(self, event=None):
        if  self.img is None:
            return
        im = self.img.copy()
        data = self.worker.source.copy()
        w_data = self.worker.im_src.copy()
        if self.savePath is None:
            dialog = wx.FileDialog ( None, wildcard = "PNG|*.png|Data| *.dat", style = wx.SAVE|wx.FD_CHANGE_DIR  )
            if dialog.ShowModal() == wx.ID_OK:
                self.savePath = dialog.GetDirectory()
                dialog.Destroy()
            else:
                dialog.Destroy()
                return

        dirlist = os.listdir(self.savePath)
        cnt =-1
        for name in dirlist:
            s = re.match('^'+IMG_PREFIX+'(\d\d\d)\.',name)
            if s is not None:
                cnt = max(cnt, int(s.group(1)))

        name = IMG_PREFIX+'{:03d}'.format(cnt+1)

        try:
            image = wx.Image(*(im.shape[:2]+(bytearray(im.transpose([1,0,2]).tostring()),)))
        except:
            image = wx.ImageFromData(*(im.shape[:2]+(bytearray(im.transpose([1,0,2]).tostring()),)))
        image.SaveFile(os.path.join(self.savePath, name+".PNG"),wx.BITMAP_TYPE_PNG)

    ##                    x = np.transpose(self.worker.data).copy()
        x = w_data
        comment = "Processed image data"
        with open(os.path.join(self.savePath, name+".dat"),"w") as out:
            out.write('# '+comment+'\n')
            out.write('# %dx%d\n' % x.shape)
            for i in range(x.shape[1]):
                out.write(",\t".join("%3d" % a for a in x[:,i])+"\n")
                    
    ##                    x = np.transpose(self.data).copy()
        x = data
        comment = "Original image data"
        with open(os.path.join(self.savePath, name+".orig.dat"),"w") as out:
            out.write('# '+comment+'\n')
            out.write('# %dx%d\n' % x.shape)
            for i in range(x.shape[1]):
                out.write(",\t".join("%.5f" % a for a in x[:,i])+"\n")



    def OnWriteVideo(self, event=None):
        fps = int(self.fps)
        if self.videoWriter is not None:
            self.writevideo.SetLabel("Save video...")
            self.status.SetStatusText( "",1)
            self.saveimg.Enable()
            self.startstop.Enable()
            del self.videoWriter
            self.videoWriter = None
        else:
            dialog = wx.FileDialog ( None, wildcard = "MPEG4 P2 Video (*.avi)|*.avi|Uncompressed AVI (*.avi)|*.avi", style = wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT  )
            if dialog.ShowModal() == wx.ID_OK:
                if min(self.img.shape[0:2]) < MIN_VIDEO_OUT:
                    k = float(MIN_VIDEO_OUT)/min(self.img.shape[0:2])
##                    self.videoout_size =  (int(k * self.img.shape[1]),int(k * self.img.shape[0]))
                    self.videoout_size =  (int(k * self.img.shape[0]),int(k * self.img.shape[1]))
                    frame_size = self.videoout_size
                else:
                    self.videoout_size = None
##                    frame_size = self.img.shape[1::-1]
                    frame_size = self.img.shape[:2]
                path = os.path.abspath(dialog.GetPath())
                if dialog.GetFilterIndex()==0:
                    try:
                        fcc =cv2.VideoWriter_fourcc('X','V','I','D')
                    except:
                        fcc =cv2.cv.CV_FOURCC('X','V','I','D')
                else:
                    fcc = 0
                    
                try:
                    self.videoWriter = cv2.VideoWriter(filename = path,fourcc = fcc, fps = fps if fps > 1 else 1, frameSize = frame_size, isColor=1)
                except:
                    self.videoWriter = cv2.VideoWriter(filename = path.encode("utf-8"),fourcc = fcc, fps = fps if fps > 1 else 1, frameSize = frame_size, isColor=1)

                self.writevideo.SetLabel("Stop")
                self.status.SetStatusText( "Recording videostream",1)
                self.saveimg.Disable()
                self.startstop.Disable()
            dialog.Destroy()

    def OnUpdate(self,event=None):
#        self.image.Update(self.worker.img)
        
        self.selection = self.image.GetSelection()
        p1, p2 = self.selection        
        if event is None or event.data:
            self.values = self.worker.statistics(self.selection)
            if p1 is None:
                self.statsData.SetLabel("")
                self.statsBox.SetLabel("")
            elif p1 == p2:
                self.statsData.SetLabel("%(mean).3f" % self.values)
                self.statsBox.SetLabel("Selection: {}, {}".format(*p1))
            else:
                self.statsData.SetLabel(("Min: %(min).3f"+4*" "+"Max: %(max).3f"+4*" "+"Average: %(mean).3e") % self.values )

                self.statsBox.SetLabel("Selection: {0}:{2}, {1}:{3}".format(*(p1+p2)))
            self.status.SetStatusText("frame #%d" % self.counter,2)
            self.status.SetStatusText( "%.2f" % self.fps +" fps",3)
        else:
            if p1 is None:
                self.statsData.SetLabel("")
                self.statsBox.SetLabel("")
            elif p1 == p2:
                self.statsData.SetLabel("")
                self.statsBox.SetLabel("Selection: {}, {}".format(*p1))
            else:
                self.statsData.SetLabel(("Min: %(min).3f"+4*" "+"Max: %(max).3f"+4*" "+"Average: %(mean).3e") % {'min':0,'max':0,'mean':0} )

                self.statsBox.SetLabel("Selection: {0}:{2}, {1}:{3}".format(*(p1+p2)))


        
    def OnResult(self, data):
        """Show Result status."""

        if self._stopme:
            wx.CallAfter(self._OnStop)
            self._stopme = False
            return True
        if data is None:
            return
        
        self.data = data.copy()
        self.img = self.worker.makeImg(data)
        self.counter += 1
        
        newtime = time.perf_counter()
        if len(self.fps_list)<11 :
            self.fps_list.append( 1./(newtime-self.lastTime) )
            self.fps = sorted(self.fps_list)[(len(self.fps_list)-1)//2]
        else:
            self.fps_list.append( 1./(newtime-self.lastTime) )
            self.fps_list.pop(0)
            self.fps = sorted(self.fps_list)[(len(self.fps_list)-1)//2]
##        try:        
##            wx.MutexGuiEnter()
##            self.image.Update(self.img.copy())
##            wx.MutexGuiLeave()
        wx.CallAfter(self.image.Update,self.img.copy())
##        except:
##            self.image.Update(self.img.copy())
        
        wx.PostEvent(self, UpdateEvent(True))
        
        self.lastTime = newtime
        if self.videoWriter is not None and self.videoWriter.isOpened():
##            im = self.img.copy()
            im = self.img.transpose([1,0,2]).copy()
            if self.videoout_size is not None:
                self.videoout_frame = cv2.resize(im, self.videoout_size,interpolation=cv2.INTER_NEAREST)                
            else:
                self.videoout_frame = im
            cv2.cvtColor(self.videoout_frame, cv2.COLOR_RGB2BGR, self.videoout_frame)
            self.videoWriter.write(self.videoout_frame)

    
    def OnError(self, event=None, critical=True):
        """Hardware error"""
            
        dialog = wx.MessageDialog(None, "Runtime Error:\n"+str(event)+("\n\nApplication will be closed." if critical else ""), 'Terasense View', wx.OK | wx.ICON_ERROR)
        dialog.ShowModal()
        dialog.Destroy()
        if critical:
            self.Close()
        return
        
    

        

    def OnSize(self, event=None):
        """ Process resize event - keep aspect ratio"""
        
        size = self.GetSize()
        if event :
            event.Skip()
            return
            if abs(self.currentSize[0]-size[0])>abs(self.currentSize[1]-size[1]) :
                newsize = (size[0], int(float(size[0])*self.currentSize[1]/self.currentSize[0]) )
            else:
                newsize = (int(float(size[1])*self.currentSize[0]/self.currentSize[1]), size[1] )
            self.SetSize(newsize)
            self.currentSize = newsize
            event.Skip()


# Main wxPython App
class MainApp(wx.App):
    """Class Main App."""
    def OnInit(self):
        """Init Main App."""
        self.name = "TeraFAST"
        self.instance = wx.SingleInstanceChecker(self.name)
        
        if not commandline_options.multihead:
            if self.instance.IsAnotherRunning():
                wx.MessageBox("ERROR - Another instance is running", "TeraFAST Viewer")
                return False

        
        self.frame = MainFrame(None, -1,style= wx.DEFAULT_FRAME_STYLE, sn = commandline_options.serial )

        
        self.frame.Show(True)
        self.SetTopWindow(self.frame)
        return True

if __name__ == '__main__':
    version = '4.2.3'#version - do not delete this comment!
    parser = ArgumentParser()
    parser.add_argument('--version',action='version',version=version)
    parser.add_argument("-m", "--multi",
                      action="store_true", dest="multihead", help="allow multiple instances for multihead operation", default=False)
    parser.add_argument("-s", "--serial",
                      action="store", type=str, dest="serial", help="try to open device with the given serial number", default='')
    parser.add_argument("-l", "--list",
                      action="store_true", dest="list", help="list serial numbers of available devices and exit", default=False)
    parser.add_argument("-t", "--test",
                      action="store_true", dest="testmode", help="run in a test mode, without hardware", default=False)    

    commandline_options = parser.parse_args()

    if commandline_options.list:
        from teraFAST import iface as iface
        de = iface.deviceEnumerator()
        for s in de.getDevList():
            print(s)
    else:
        if commandline_options.testmode:
            from teraFAST import iface as iface
            iface.test = True
    
        app = MainApp(0)
        
        app.MainLoop()
        wx.Exit()
