import struct
import numpy
import teraFAST.lib.ok_ctl as ok_ctl
from teraFAST.lib.ok_ctl import ControllerError
from teraFAST.lib.bunch import Bunch

class EP(object):
	"""Interface endpoints related definitions"""
	# endpoint addresses
	cfg    = 0
	cfg2   = 1
	itime  = 2
	period = 3
	aux    = 4
	psdiv  = 5	   
	vers   = 0x20
	debug  = 0x30
	debugx = 0x31
	trig   = 0x40
	data   = 0xa0

	# start trigger (bit index)
	trig_start = 0

	# Status bits
	sta_start_byte   = 0xff
	sta_fsync2_bit   = 0x100
	sta_chan_err_off = 16  # channel error bits offset
	sta_bits_off     = 24  # status bits offset
	sta_op_err   = 1 << 29 # fatal operation error
	sta_sync_err = 1 << 30 # synchronization error
	sta_sof      = 1 << 31 # start of the frame flag
	sta_bytes    = 4 # start of frame header bytes
	# Version 2 status bits (with position code)
	sta2_fsync2_bit   = 1
	sta2_op_err       = 1 << 5 # fatal operation error
	sta2_sync_err     = 1 << 6 # synchronization error
	sta2_sof          = 1 << 7 # start of the frame flag
	sta2_chan_err_off = 8 # channel error bits offset
	sta2_bytes        = 6 # start of frame header bytes (last 4 are position code)
	
	pcode_cnt_bits = 24
	pcode_cnt_mask = (1<<pcode_cnt_bits) - 1

	# the number of frames that can fit in internal buffer
	buffer_frames = 32

	# formatting helpers
	@staticmethod
	def unpack_vers(v):
		"""Returns (version, channels, emulated chain) tuple"""
		return (v >> 10, v & 0xf, (v >> 4) & 0xf)

	@staticmethod
	def pack_cfg(chain, irst_time, time10mult = False, emu = False, use_leds = False):
		"""Make config word"""
		assert 0 < chain <= 64
		assert 0 < irst_time <= 32
		v = chain | ((irst_time-1) << 8)
		if time10mult: v |= 1 << 13
		if emu:        v |= 1 << 14
		if use_leds:   v |= 1 << 15
		return v

	@staticmethod
	def pack_cfg2(ps_deb, ps_emu_per, ps_edge, ps_start_en, ps_emu_en, rd_clk_div = 1, fsync2_en = False):
		assert 0 < rd_clk_div <= 16
		v = ps_deb | (ps_emu_per << 4) | ((rd_clk_div - 1) << 11)
		if ps_edge:     v |= (1 << 8)
		if ps_start_en: v |= (1 << 9)
		if ps_emu_en:   v |= (1 << 10)
		if fsync2_en:   v |= (1 << 15)						
		return v

class ADSMC(ok_ctl.ok_ctl):
	"""ADS8568 multichannel controller"""
	def __init__(self, cfg):
		ok_ctl.ok_ctl.__init__(self, cfg)

	def init(self, force_load=False, port=''):
		"""Initialize firmware"""
		if not self.init_controller(force_load=force_load, port=port):
			return False
		self.dev.UpdateWireOuts()
		v = self.dev.GetWireOutValue(EP.vers)
		version, channels, emu_chain = EP.unpack_vers(v)
		if not version or channels > 8:
			raise ControllerError('unexpected version code %#x' % v)
		self.ver_info = (version, channels, emu_chain)
		return True

	def fw_version(self):
		return self.ver_info[0]

	def fw_channels(self):
		return self.ver_info[1]

	def fw_emulated_chain(self):
		return self.ver_info[2]

	def info(self):
		"""Returns info string"""
		return 'OpalKelly ADS8568 %d channels controller %s v.%d, %dx%d configured (%d emulated)' % (
				self.fw_channels(), self.get_id(), self.fw_version(), self.cfg.channels, self.cfg.chain_len, self.fw_emulated_chain()
			)

	def configure(self, emu_chain=None, emu_psync_per=None):
		"""Setup controller configuration"""
		assert 0 < self.cfg.chain_len <= 64
		assert 0 < self.cfg.itime  <= (1 << 16)
		assert 0 < self.cfg.period <= (1 << 16)
		assert emu_chain is None or 0 < emu_chain <= self.fw_emulated_chain()
		self.last_cfg = Bunch(
				chain      = self.cfg.chain_len if emu_chain is None else emu_chain,
				irst_time  = self.cfg.irst_time,
				time10mult = self.cfg.time10mult,
				emu        = emu_chain is not None,
				use_leds   = self.cfg.use_leds if emu_chain is None else True
		)
		cfg = EP.pack_cfg(**self.last_cfg)
		if	self.dev.SetWireInValue(EP.cfg, cfg) != self.dev.NoError or \
			self.dev.SetWireInValue(EP.itime,  self.cfg.itime  - 1) != self.dev.NoError or \
			self.dev.SetWireInValue(EP.period, self.cfg.period - 1) != self.dev.NoError:
			raise ControllerError('failed to setup configuration')
		if self.fw_version() >= 2:
			cfg2 = EP.pack_cfg2(
				ps_deb      = self.cfg.psync_debounce,
				ps_emu_per  = emu_psync_per if emu_psync_per is not None else 0,
				ps_edge     = self.cfg.psync_edge,
				ps_start_en = self.cfg.psync_start_en,
				ps_emu_en   = emu_psync_per is not None,
				rd_clk_div  = self.cfg.rd_clk_div,
                                fsync2_en = self.cfg.fsync2_en
			)
			if	self.dev.SetWireInValue(EP.cfg2, cfg2) != self.dev.NoError:
				raise ControllerError('failed to setup configuration')
		if self.fw_version() >= 3:
			psync_div = self.cfg.psync_div
			if psync_div > 1:
				psync_div -= 1
			else:
				psync_div = 0
			if self.dev.SetWireInValue(EP.psdiv, psync_div) != self.dev.NoError:
				raise ControllerError('failed to setup psync divider')
		self.dev.UpdateWireIns()

	def set_aux(self, v):
		"""Set auxiliary configuration. Currently 3 low level bits are going to general purpose outpus"""
		if self.dev.SetWireInValue(EP.aux, v) != self.dev.NoError:
			raise ControllerError('failed to write auxiliary configuration word')
		self.dev.UpdateWireIns()

	def get_debug_info(self):
		"""Read debug info register"""
		self.dev.UpdateWireOuts()
		return (self.dev.GetWireOutValue(EP.debugx) << 16) | self.dev.GetWireOutValue(EP.debug)


	def acquire_raw(self, cb):
		"""
		Run acquisition in infinite loop passing sample index, status word and data buffer to the supplied callback.
		If callback returns True, the loop terminates.
		"""
		if self.dev.ActivateTriggerIn(EP.trig, EP.trig_start) != self.dev.NoError:
			raise ControllerError('failed to start acquisition')
		data_sz, I = self.fw_channels() * self.last_cfg.chain * 8 * 2, [0]
		ver = self.fw_version()
		if ver < 2:
			hdr_bytes = EP.sta_bytes
			st_byte = EP.sta_start_byte
			def proxy(data):
				status, i = struct.unpack('I', data[:EP.sta_bytes])[0], I[0]
				if (status & 0xff != EP.sta_start_byte) or not (status & EP.sta_sof) or (status & EP.sta_op_err):
					raise ControllerError('controller reported fatal acquisition error (status %#x)' % status)
				assert (not (status & EP.sta_fsync2_bit)) == (not (I[0] & 1))
				res = cb(i, (status, 0), data[EP.sta_bytes:])
				I[0] += 1
				return res
		else:
			hdr_bytes = EP.sta2_bytes
			st_byte = None
			def proxy(data):
				# status, pcode = struct.unpack('<HI', str(data[:EP.sta2_bytes]))
				status, pcode = struct.unpack('<HI', data[:EP.sta2_bytes])
				i = I[0]
				if not (status & EP.sta2_sof) or (status & EP.sta2_op_err):
					raise ControllerError('controller v.%u reported fatal acquisition error (status %#x)' % (ver, status))
				assert (not (status & EP.sta2_fsync2_bit)) == (not (I[0] & 1))
				status_bits, status_chan = status & 0xff, (status >> EP.sta2_chan_err_off) & 0xff
				# Convert status to v.1 format for compatibility
				res = cb(i, (EP.sta_start_byte | (status_chan << EP.sta_chan_err_off) | (status_bits << EP.sta_bits_off), pcode), data[EP.sta2_bytes:])
				I[0] += 1
				return res

		self.read_chunk_stream(
				EP.data, hdr_bytes + data_sz, grp=2*EP.buffer_frames,
				cb=proxy, msec_tout=2000, start_byte=st_byte
			)

	def acquire_data(self, channels, cb, idx=None, dt=numpy.int16):
		"""
		Run acquisition in infinite loop passing sample index, and data array to the supplied callback.
		If callback retrurns True, the loop terminates.
		"""
		if idx is None:
			# Create default index
			sample_group = self.fw_channels() * 4
			idx = numpy.array([((i//2) if i % 2 == 0 else (i//2) + sample_group) + j * 2 * sample_group
					for j in range(self.last_cfg.chain)
						for i in range(channels * 8)
				])
		def proxy(i, status, data):
			
			if ((status[0] >> EP.sta_chan_err_off) & ((1 << channels) - 1)):
                                pass
                                raise ControllerError('controller reported channel error (status %#x)' % status[0])
			d = numpy.frombuffer(data, dt)
			return cb(i, status, numpy.take(d,idx))

		self.acquire_raw(proxy)

