#
# The module provides API for accessing M25P32 flash memory present on opal kelly boards.
# It accompany the HDL code that must be incorporated onto your project.
#

from opal_kelly import *
from collections import namedtuple
import time
import struct
import zlib
import sys

try:
        from .flashapi6_data import  flashapi6_compressed
except ImportError:
        flashapi6_compressed = None
        
try:
        from .flashapi7_data import flashapi7_compressed        
except ImportError:
        flashapi7_compressed = None

def get_flashapi(version):        
        if version == OK_PRODUCT_XEM6001:
                try:
                        return zlib.decompress(flashapi6_compressed)
                except:
                        return None
        elif version == OK_PRODUCT_XEM7001A15 or version == OK_PRODUCT_XEM7001A35:
                try:
                        return zlib.decompress(flashapi7_compressed)
                except:
                        return None        
        else:
                return None

FW_ID = 0  # id of firmware blob in the flash memory
FW_INFO = "firmware"

CALIBRATION_ID  = 100   # id of configuration blob in flash memory
CALIBRATION_INFO = "calibration"
CALIBRATION_MAGIC_STRING = "TS_CAL"

page_sz   = 256
sector_sz = 64*1024
total_sz  = 4*1024*1024
total_pages    = total_sz // page_sz
total_sectors  = total_sz // sector_sz

# endpoints for interfacing to HDL
ep_address = 0x1f
ep_command = 0x5f
ep_status  = 0x3f
ep_buff_rd = 0xbf
ep_buff_wr = 0x9f

# commands (bit index for [ep_command])
cmd_rd_status  = 0
cmd_wr_enable  = 1
cmd_bulk_erase = 2
cmd_sect_erase = 3
cmd_rd_data    = 4
cmd_wr_page    = 5

# status bits
sta_version_shift = 13
sta_mask = (1 << 8) - 1
sta_wr_inprogr = 1 << 0
sta_wr_enabled = 1 << 1
sta_tx_busy    = 1 << 8

# sector header provides the convenient mean of organizing your data
sector_hdr_magic = 0xa365477d
sector_hdr_fmt = 'IIIIII80s' # magic followed by tag
sector_hdr_sz  = struct.calcsize(sector_hdr_fmt)
max_sector_data_sz = sector_sz - sector_hdr_sz

# The maximum length of human readable information string
max_sector_info_sz = 80

# If the following flag is set the sector contains compressed data
sector_fl_compressed = 1
sector_compression_level = 9

SectorTag = namedtuple('SectorTag', ('key', 'version', 'flags', 'crc', 'size', 'info'))
FreeSectorTag = SectorTag(key=None, version=None, flags=None, crc=None, size=0, info=b'free sector')

## blob id is converted into key by left-shifting by FLASH_ID_SHIFT */
FLASH_ID_SHIFT = 8 


def crc32(s, crc=0):
        """Calculate crc32 checksum"""
        return zlib.crc32(s, crc) & 0xffffffff

def get_version(dev):
        """Query firmware version"""
        dev.UpdateWireOuts()
        sta = dev.GetWireOutValue(ep_status)
        return sta >> sta_version_shift

def api_present(dev):
        """Returns True in case flash API is present in the firmware"""
        return get_version(dev) == 1

def send_command(dev, cmd):
        """Send specified command and returns response byte (last byre received from flash)"""
        res = dev.ActivateTriggerIn(ep_command, cmd)
        if res != dev.NoError:
                raise RuntimeError('failed to read flash status')
        while True:
                dev.UpdateWireOuts()
                sta = dev.GetWireOutValue(ep_status)
                if not (sta & sta_tx_busy):
                        return sta

def read_status(dev):
        """Read flash status byte"""
        return sta_mask & send_command(dev, cmd_rd_status)

def wait_wr_completion(dev, tout):
        """Wait flash operation completion by polling status byte"""
        start_ts = None
        while True:
                if not (read_status(dev) & sta_wr_inprogr):
                        break
                if start_ts is None:
                        start_ts = time.monotonic()
                elif time.monotonic() > start_ts + tout:
                        raise RuntimeError('flash write completion timeout')
                time.sleep(tout / 100.)

def erase_all(dev, wait=True):
        """Erase entire flash. Usually takes ~25 sec to complete."""
        send_command(dev, cmd_wr_enable)
        send_command(dev, cmd_bulk_erase)
        if wait:
                wait_wr_completion(dev, 100.)

def write_buff(dev, data):
        """Write data to page buffer implemented in HDL"""
        sz = len(data)
        assert sz <= page_sz
        if sz % 2:
                dat = data.copy()
                dat.append(0)
                sz += 1
                res = dev.WriteToPipeIn(ep_buff_wr, bytearray(dat))
        else:
                res = dev.WriteToPipeIn(ep_buff_wr, bytearray(data))
        if res != sz:
                raise RuntimeError('failed to write flash buffer: %d bytes, %d bytes written' % (sz, res))

def read_buff(dev, sz):
        """Read data from page buffer implemented in HDL"""
        assert sz <= page_sz
        sz_ = sz
        if sz_ % 2:
                sz_ += 1
        buff = bytearray(sz_)
        cnt = dev.ReadFromPipeOut(ep_buff_rd, buff)
        if cnt != sz_:
                raise RuntimeError('failed to read flash buffer: %d bytes requested, %d bytes read' % (sz_, cnt))
        return buff[0:sz]

def set_address(dev, addr):
        """Set start address for flash operation. The low order 8 bits are assumed to be all 0."""
        assert addr % page_sz == 0
        res = dev.SetWireInValue(ep_address, addr // page_sz)
        if res != dev.NoError:
                raise RuntimeError('failed to set flash page address')
        dev.UpdateWireIns()

def read_page(dev, addr, sz):
        """Read up to the page of flash data"""
        assert addr % page_sz == 0
        assert sz <= page_sz
        set_address(dev, addr)
        send_command(dev, cmd_rd_data)
        return read_buff(dev, sz)

def read_data(dev, addr, sz):
        """Read arbitrary amount of flash data at arbitrary offset"""
        align = addr % page_sz
        addr -= align
        sz   += align
        buff = bytearray([])
        while sz:
                chunk_sz = min(sz, page_sz)
                data = read_page(dev, addr, chunk_sz)
                buff += data[align:]
                sz   -= chunk_sz
                addr += chunk_sz
                align = 0
        return buff

def write_page(dev, addr, buff):
        """Write up to the page of data to the flash"""
        assert addr % page_sz == 0
        if len(buff) < page_sz:
                buff+= b'\xff' * (page_sz - len(buff))
        set_address(dev, addr)
        send_command(dev, cmd_wr_enable)
        write_buff(dev, buff)
        send_command(dev, cmd_wr_page)
        wait_wr_completion(dev, .01)

def write_data(dev, addr, buff):
        """Write page aligned data to flash"""
        assert addr % page_sz == 0
        sz, off = len(buff), 0
        while sz:
                chunk_sz = min(sz, page_sz)
                write_page(dev, addr, buff[off:off+chunk_sz])
                sz   -= chunk_sz
                off  += chunk_sz
                addr += chunk_sz

def erase_sector(dev, sn, wait=True):
        """Erase flash sector given its number"""
        set_address(dev, sn * sector_sz)
        send_command(dev, cmd_wr_enable)
        send_command(dev, cmd_sect_erase)
        if wait:
                wait_wr_completion(dev, 5.)

def sector_tag_from_buff(buff):
        """Create sector tag from header buffer"""
        assert len(buff) == sector_hdr_sz
        magic, key, version, flags, crc, size, info = struct.unpack(sector_hdr_fmt, buff)
        if magic == sector_hdr_magic and 0 < size <= max_sector_data_sz:
                return SectorTag(key, version, flags, crc, size, info.rstrip(b'\0'))
        else:
                return FreeSectorTag

def read_sector_tag(dev, sn):
        """Read sector header"""
        assert 0 <= sn < total_sectors
        hdr = read_page(dev, sn * sector_sz, sector_hdr_sz)
        return sector_tag_from_buff(hdr)

def get_sector_map(dev):
        """Read all sector headers and return them in array"""
        return [read_sector_tag(dev, sn) for sn in range(total_sectors)]

def find_sector_by_key(dev, key):
        """Find sector given its key and returns its number"""
        for sn in range(total_sectors):
                tag = read_sector_tag(dev, sn)
                if key == tag.key :
                        return sn, tag
        return None, None

def find_free_sector(dev):
        """Find first free sector and returns its number"""
        sn, tag = find_sector_by_key(dev, None)
        return sn

def write_sector(dev, sn, key, data, version=0, flags=sector_fl_compressed, info=b''):
        """Write key, data to the sector given its number"""
        sz, max_data_sz = len(data), max_sector_data_sz        
        if flags & sector_fl_compressed:
                original_sz = sz
                data = zlib.compress(data, sector_compression_level)
                sz, crc = len(data), crc32(data)
                data += struct.pack("I", original_sz)
                max_data_sz -= struct.calcsize("I")
        else:
                crc = crc32(data)
        if sz > max_data_sz:
                raise RuntimeError('sector data too large')
        erase_sector(dev, sn)
        if isinstance(info,str):
                info_=bytes(info,encoding='utf8')
        else:
                info_ = bytes(info)
        hdr = struct.pack(sector_hdr_fmt, sector_hdr_magic, key, version, flags, crc, sz, info_)
        buff = hdr + data
        write_data(dev, sn * sector_sz, buff)
        buff_ = read_data(dev, sn * sector_sz, len(buff))
        if buff != buff_:
                raise RuntimeError('flash sector data verification failed')

def read_sector(dev, sn, key):
        """Read sector given its number and key which must match the one stored in flash"""
        sector_off = sn * sector_sz
        page = read_page(dev, sector_off, page_sz)
        tag = sector_tag_from_buff(page[:sector_hdr_sz])
        if key != tag.key:
                return None
        compressed = tag.flags & sector_fl_compressed
        total_sz = sector_hdr_sz + tag.size
        if compressed:
                suff_sz = struct.calcsize("I")
                total_sz += suff_sz
        if total_sz <= page_sz:
                data = page[sector_hdr_sz:total_sz]
        else:
                data = page[sector_hdr_sz:] + read_data(dev, sector_off + page_sz, total_sz - page_sz)
        if compressed:
                original_sz, = struct.unpack("I", data[-suff_sz:])
                data = data[:-suff_sz]
        if crc32(data) != tag.crc:
                raise RuntimeError('flash sector data corrupted')
        if compressed:
                data = zlib.decompress(data)
                if len(data) > original_sz:
                        raise RuntimeError('invalid sector data size')
        return data


## working with blobs



def find_blob(dev, blob_id):
        """Checks if a blob with given id is present in flash memory. Returns first sector number on success, or TS_OK_FLASH_SECTOR_INVALID if not found, or negative error code in case of error.
On success info string is populated. """

        key = (blob_id+1) << FLASH_ID_SHIFT
        sn, tag = find_sector_by_key(dev, key)
        if (sn is None):
                return sn, ""
        return sn, tag.info

def delete_blob(dev, blob_id):
        """ Deletes a blob with given id from flash memory. """

        key = (blob_id + 1) << FLASH_ID_SHIFT
        sn, _ = find_sector_by_key(dev, key)
        while (sn is not None):
                erase_sector(dev, sn, True)
                key += 1
                sn, _= find_sector_by_key(dev, key)





def write_blob(dev, blob_id, info:'bytearray', data:'bytearray'):
        """Write a blob to flash. It is identified in flash memory by id. If the id is already present data are overwritten.
Returns True on success or raise RuntimeError in case of error. """

        delete_blob(dev, blob_id)

#first four bytes are length of the data  
        blob = struct.pack("=I",len(data))
        
        compressed = zlib.compress(data)
        compressed_sz = len(compressed)

        blob += compressed
     
        p = 0
        count = 0
        remaining_sz = len(blob)
        key = (blob_id + 1) << FLASH_ID_SHIFT
        while (remaining_sz > max_sector_data_sz):
                sn = find_free_sector(dev)
                if (sn is None):
                        delete_blob(dev,blob_id) 
                        raise RuntimeError("Not enough free memory left in flash to write %d bytes"%(compressed_sz))
                write_sector(dev, sn, key + count, blob[p:p+max_sector_data_sz], info=info, flags=0)
                count +=1
                p += max_sector_data_sz
                remaining_sz -= max_sector_data_sz      
        if (remaining_sz > 0):
                sn = find_free_sector(dev)
                if (sn is None):
                        flash_delete(dev,blob_id)
                        raise RuntimeError("Not enough free memory left in flash to write %d bytes"%(compressed_sz))
                        
                write_sector(dev, sn, key + count, blob[p:], info=info, flags=0)
                
                count+=1
        return True


def read_blob(dev, blob_id):
        """Read a blob from flash. It is identified in flash memory by id. Data buffer is allocated by the function, so it need to be released by user.
Returns   data, info  on success, (None, None) if no data with such id have been found, or raises RuntimeError in case of error. """

        key = (blob_id + 1) << FLASH_ID_SHIFT

        sn, info = find_blob(dev,blob_id)
        if (sn is None):
                
## "Data not found
                return None, None
        data = bytearray([])
        count = 0
        sd = read_sector(dev, sn, key)
        if (sd is None):
                raise RuntimeError("Error reading sector %d" % sn)
                
        count+=1

        data.extend(sd)
        
        sn, _ = find_sector_by_key(dev, key + count)
        while (sn is not None):
                sd = read_sector(dev, sn, key + count)
                count +=1
                data.extend(sd)
                sn, _ = find_sector_by_key(dev, key + count)

## first four bytes are size
        result = zlib.decompress(data[4:])
        if struct.unpack("=I",data[:4])[0] != len(result):
                raise RuntimeError("Error reading flash data: unpacked size %d does not match the header %d " % (struct.unpack("=I",data[:4])[0], len(result))  )
        return result, info

