#from builtins import str
#from builtins import object
#from past.utils import old_div
from importlib import import_module
import threading 
import numpy as np
import cv2
import zipfile, os, sys, tempfile
import io
import struct

from queue import Full, Empty
from . import ref as ref

import teraFAST.iface as iface

import sys

RAW         =   0x01
DIFFERENCE  =   0x10
DEFAULT_FLAGS = 0x0

DEFAULT_BG_FILENAMES = ["0.bg"]
DEFAULT_NORM_KEYS = ("default","recorded")
DEFAULT_NORM_FILENAMES = {"default":"norm_default.nm","recorded":"norm_recorded.nm"}

RAW_MAX = 0.7

MAX_VALUE = (1<<15)-1
MAX_FLOAT = 1e99

## binary calibration scales
SCALE_NORM = 1
SCALE_MASK = 1
SCALE_BG = 10

class AutoModeException(Exception):
    def __init__(self, stop = True):
        self.stop = stop
    

class deviceEnumerator(iface.deviceEnumerator):
    def __init__(self):
        iface.deviceEnumerator.__init__(self)
        
class progress(object):
    def __init__(self,total=100,callback=None):
        """Helper class for managing progress of a process.
    total - length of the process in arbitrary units (default 100)
    callback - optional function to call back on change with percentage of the total as an argument."
"""
        self.current = 0.0
        self.total = float(total)
        self.progress = self.current/self.total
        if callable(callback):
            self.callback = callback
            self.callback(0)
        else:
            self.callback = None
    def increase(self,value=1):
        """Augment progress by the value.
    value - size of the increment (default 1)
"""
        self.current += value
        if self.current > self.total:
            self.current = self.total
        self.progress = self.current/self.total
        if self.callback is not None:
            return self.callback(int(self.progress*100))
        return True, False
    
    def get_progress(self):
        """Get current progress in percents of total"""
        return int(self.progress*100)

class processor(object):
    """Main class for data acquisition and processing."""


  
    def __init__(self,rows = 1, threaded=True, config=None, defaults=None, flags = DEFAULT_FLAGS, sn=''):
        """Creates interface to the hardware and setups acquisition and data processing.
    Parameters:
        rows        -   number of rows in the returned image
        config      -   path to a configuration file containing background and normalization data (default: None)
        defaults    -   path to a defaults file (default: None)
        flags       -   data processing flags (default: DEFAULT_FLAGS)
        sn          -   serial number string of the device to open (default: empty string=the first available)

    Data processing flags:
        DIFFERENCE  -   turn on differential signal between consecutive frames
        DEFAULT_FLAGS = 0x0
"""


        if defaults is None:
            defaults = import_module(".config","teraFAST.config")
        if config is None:
            try:
                config = os.path.join(os.path.dirname(__file__),defaults.configPath)
            except:
                config = ""

        self.threaded = True

        
##        self.interface = iface.instance
        self.interface = iface.interface()
        
        self.interface.init(sn)
        self.sn = self.interface.get_id()
        
        self.rows = int(rows) if int(rows)>1 else 1        

        self.X_SIZE = self.interface.X_SIZE
        self.Y_SIZE = self.rows

        self.buffer =  np.empty((self.X_SIZE,self.Y_SIZE),dtype=float)
        self.result =  np.empty((self.X_SIZE,self.Y_SIZE),dtype=float)

        self.processFlags = flags

        
        self._want_setrate = None

        self._callback = None

        self.currentNorm = None
        self.currentBG = None
        

        #import config.proc as defaults




        

        if self.threaded:
            self.datalock = threading.Lock()
            self.ready = threading.Event()
            self.proc_thread = None
            self._stop_processing = False

        self.bg = self.bgNone = np.zeros((self.X_SIZE,self.Y_SIZE),dtype=float)
        self.norm = np.ones((self.X_SIZE,self.Y_SIZE),dtype=float)*MAX_VALUE
        self.normNone = np.ones((self.X_SIZE,self.Y_SIZE),dtype=float)*MAX_VALUE
        self.normSrc = np.ones((self.X_SIZE,self.Y_SIZE),dtype=float)*MAX_VALUE
        
        self.mask = 1e99*np.ones((self.X_SIZE,self.Y_SIZE),dtype=float)
        self.maskNone = 1e99*np.ones((self.X_SIZE,self.Y_SIZE),dtype=float)
        self.odd_frame = np.zeros((self.X_SIZE,self.Y_SIZE),dtype=float)


        try:
            self.threshold = defaults.threshold # limiting signal/noise ratio during calibration for mask
        except AttributeError:
            self.threshold = 10

        try:
            self.recordingCount = defaults.recordingCount # limiting signal/noise ratio during calibration for mask
        except AttributeError:
            self.recordingCount = 30


        configZipName = None
        if os.path.isfile(config):
            configZipName = config
##        elif os.path.isfile(os.path.join(config,self.interface.get_id()+".zip")):
##            configZipName = os.path.join(config,self.interface.get_id()+".zip")
        elif self.interface.ctl is not None and self.interface.ctl.cal_data is not None:
            try:
                print("loading from memory", len(self.interface.ctl.cal_data),file=sys.stderr)
                self.LoadConfigFromMemory(self.interface.ctl.cal_data)
            except Exception as e:
                print("Failed to load config from flash ROM:" + str(e), file=sys.stderr)
                configZipName = os.path.join(os.path.dirname(__file__),"data/"+self.interface.get_id()+".zip")        
        else:
            configZipName = os.path.join(os.path.dirname(__file__),"data/"+self.interface.get_id()+".zip")

        if configZipName is not None:
            try: 
                self.LoadConfig(configZipName)
            except FileNotFoundError as e:
                print("Failed to load config "+configZipName+" :"+str(e),file=sys.stderr)
                self.MakeNullBGList()
                self.MakeNullNormList()
                self.SelectNorm('none')



        try:
            if (defaults.difference):
                self.processFlags ^= DIFFERENCE
        except AttributeError:
            pass
        
        self.stored = self.bgNone 
        self._reset = True

        self.parity = True
        
        try:
            self.SetRate(defaults.rate)
        except AttributeError:
            self.SetRate(self.GetRateRange()[1])
        
                        
          
    def processing(self,queue,callback=None,errorcallback=None):
        """Target for data processing thread"""
        call = callable(callback)
        while not self._stop_processing:
            
            if self._want_setrate is not None:
                self.reconfigure()
                self.interface.reset()

            self.interface.lock.acquire()
            try:
                data = queue.get(timeout=0.5)
            except Empty:
                if not self._stop_processing and call:
                    self._stop_processing = callback(None)
                continue
            finally:
                self.interface.lock.release()
            
            if self._stop_processing:
                break

            if data.shape[1] != self.rows:
                self._setrows(data.shape[1])

##            if (self.processFlags & DIFFERENCE):
##                self.parity = not self.parity
##                temp = data.copy()             
##                if self.parity:
##                    np.subtract(data,self.odd_frame,data)
##                else:
##                    np.subtract(self.odd_frame,data,data)
##                self.odd_frame = temp


            data = self.process(data.astype(float))  # dtype changes from int16 to float
            self.datalock.acquire()
            self.result = data.copy()
            self.ready.set()
            self.datalock.release()
            
            if call and not self._stop_processing:
                self._stop_processing = callback(self.result)
            
        self.interface.stop()
        self._stop_processing = False
        

    def start(self,callback=None,errorcallback=None,resume=False):
        """Start multi-threaded acquisition and processing.
    Parameters:
        callback    -   a target to be called on each acquisition cycle as callback(data), where data is a numpy.ndarray containing acquired data [0,1] (default: None)
        errorcallback - a target to be called in case of an error as errorcallback(error, critical), where "error" is an instance of Exception, and "critical" is a boolean (default: None)
        resume  -   if True and callback is None the value of callback from previous call for the function will be used (default: False).
"""
        if callable(callback):
            self._callback = callback
        elif resume and callable(self._callback):
            callback = self._callback
        else:
            self._callback = None

        self.odd_frame = np.zeros((self.X_SIZE,self.Y_SIZE),dtype=float)
        self.parity = True
        
        if self.threaded:
            if self.proc_thread is None or not self.proc_thread.is_alive():
                self.interface.reset()
                try:
                    queue = self.interface.start(self.rows)
                except RuntimeError as e:
                    if callable(errorcallback):
                        errorcallback(e)
                else:
                    self.proc_thread = threading.Thread(target=self.processing,args=(queue,self._callback,errorcallback))
                    self.ready.clear()
                    self.proc_thread.start()

            

    def stop(self, join=True):
        """Stop multi-threaded acquisition and processing.
    Parameters:
        join    -   if True, it will try to join processing thread. 
"""
        if self.threaded:
            self._stop_processing = True
            self.ready.set()
            if join and self.proc_thread is not None and self.proc_thread.is_alive():                
                self.proc_thread.join(5.0)
            if join:
                self.interface.reset()

               


    
    def read(self):
        """Reads processed data. Returns numpy.ndarray with acquired data or None if acquisition is not running."""
        if self.threaded and self.proc_thread is not None and self.proc_thread.is_alive():
            if self.ready.wait(30.0):
                self.datalock.acquire()
                data = self.result.copy()
                self.ready.clear()
                self.datalock.release()

                return data

        return None

    def read_raw(self):
        """Reads unprocessed data. Returns numpy.ndarray with acquired data or None if acquisition is not running."""
        data = self.interface.read() 
        return data

    def readBGData(self):
        """ Returns raw data for background recording (no processing). """
        data = self.interface.read()
        if data is None:
            return None
        else:
            return data.astype(float)        



    def readNormData(self):
        """ Returns raw data for normalization recording (background substracted). """
        data = self.interface.read()
        if data is None:
            return None
        else:
            if not (self.processFlags & DIFFERENCE) :
                data = self.SubstractBG(data.astype(float))
            return data.astype(float)

    def _setrows(self,rows):
##        flag = self.threaded and self.proc_thread is not None and self.proc_thread.is_alive()
##        if flag:
##            self.stop()
                
        self.rows = int(rows) if int(rows)>1 else 1
        self.Y_SIZE = self.rows
        
        self.buffer =  np.empty((self.X_SIZE,self.Y_SIZE),dtype=float)
        self.result =  np.empty((self.X_SIZE,self.Y_SIZE),dtype=float)
#        self.temp =  np.empty((self.X_SIZE,self.Y_SIZE),dtype=float)

        self.bgNone = np.zeros((self.X_SIZE,self.Y_SIZE),dtype=float)
        self.bg = np.zeros((self.X_SIZE,self.Y_SIZE),dtype=float)
        self.normNone = self.normSrc = np.ones((self.X_SIZE,self.Y_SIZE),dtype=float)*MAX_VALUE
        self.norm = np.ones((self.X_SIZE,self.Y_SIZE),dtype=float)*MAX_VALUE
        self.mask = self.maskNone = 1e99*np.ones((self.X_SIZE,self.Y_SIZE),dtype=float)
        self.odd_frame = np.zeros((self.X_SIZE,self.Y_SIZE),dtype=float)
        self.parity = False
        
        self.SelectBG(0)
        self.SelectNorm(self.GetSelectedNorm())
        
##        if flag:
##            self.start(resume=True)

    def SetFrameLength(self, x):
        x = int(x) if int(x)>1 else 1
        if self.threaded and self.proc_thread is not None and self.proc_thread.is_alive():
            self.interface.SetFrameLength(x)
        else:
            self._setrows(x)


    def SetSyncOut(self, on):
        """Turns sync output signal on or off."""                    
        return self.interface.SetSyncOut(on)
    
    def GetSyncOut(self):
        """Returns current state of sync output."""
        return self.interface.GetSyncOut()


    def SetExternalSync(self, x, edge=None, divider = None):
        """Turns external sync mode on."""
        if edge is not None:
            self.interface.SetEdge(bool(edge))
        if divider is not None:
            self.interface.SetSyncDivider(divider)
        return self.interface.ExternalSync(x)
    
    def GetExternalSync(self):
        """Returns current external sync setting."""
        return self.interface.GetExternalSync()
    

    def SetEdge(self, rise):
        """Sets edge of the external signal to trigger line. True = rising, False = faling.
Only has effect on next acquisition restart or turning external sync mode on."""
        return self.interface.SetEdge(rise)
    
    def GetEdge(self):
        """Returns current sync edge setting"""
        return bool(self.interface.GetEdge())

    def SetSyncDivider(self, div):
        """Sets divider for external sync signal frequency, 1 <= div <= 32768
"""
        return self.interface.SetSyncDivider(div)
    
    def GetSyncDivider(self):
        return self.interface.GetSyncDivider()


    def SetSyncHoldoff(self, holdoff):
        """Sets holdoff parameter for the external sync signal, 0 <= div <= 15.
It corresponsd to minimum 2^holdoff microseconds delay between active edges. 
"""
        return self.interface.SetSyncHoldoff(holdoff)
    
    def GetSyncHoldoff(self):
        return self.interface.GetSyncHoldoff()

    def GetID(self):
        """Returns ID string of the connected device"""
        return self.interface.get_id()
    # processing functions
    def process(self, data):
        """ Process data. """

        if not (self.processFlags & DIFFERENCE) :
            data = self.SubstractBG(data)
            
        data = self.Normalize(data)
        
        return data


    def SubstractBG(self, data):
        np.subtract(data,self.bg,data)
        return data

    def Normalize(self, data):
        data=np.true_divide(data,self.norm)
##        if self.norm is self.normNone:
        np.absolute(data,data)
        np.clip(data,0,1,data)
        return data


    #helper functions
    
    def prepareNorm(self):
        """Prepare data to be used in normalization.

        Returns:
            None
"""
        try:
            self.norm = self.normSrc.copy()
        except:
            self.norm = np.array(self.normSrc,dtype=float)

        self.norm = np.where(  self.mask >  self.threshold,self.norm, np.inf )

    def SetThreshold(self, value):
        """Set threshold value for the signal-to-noise ratio to declare pixels unusable with recorded normalization.
        Parameters:
            value - the threshold (invalid values less than 1 defaults to 10).
        Returns:
            None
"""               
            
    def SetRate(self,rate):
        """Changes current acquisition rate.
    Parameters:
        rate -   new rate value, lines/second.
    Returns:
        In case when acquisition is not running returns None for success or Exception instance for failure;
        In case of running acquisition always returns None and in case of failure errorcallback() is called by the processing thread.
""" 
        if self.threaded and self.proc_thread is not None and self.proc_thread.is_alive():
            self._want_setrate = rate
            return None
        else:
            return self._setrate(rate)


    def _setrate(self,rate):
        """Low-level (not thread-aware) change in current exposure.
    Parameters:
        exp -   new exposure value.
    Returns:
        None for success or Exception instance for failure.
"""
        error = self.interface.SetRate(rate)
##        self.SelectBG(exp)            
        self._want_setrate = None
        return error
            

    def GetRate(self):
        """Gets current acquisition rate.

    Returns:
        The acquisition rate, lines/second. The return value may differ from the one used at SetRate() due to the rounding errors."""
        return self.interface.GetRate()

    def GetRateRange(self):
        """Gets admissible rate range.

    Returns:
        (min, max)  - a tuple with minimal and maximal rate values, lines/second.
"""
        return self.interface.GetRateRange()
        
    
    def reconfigure(self):
        error = None
        if self._want_setrate is not None:
            error = self._setrate(self._want_setrate)                
            self._want_setrate = None
        return error
        


    def SetBG(self,data):
        """Sets data as current backgound data to be used in processing.
    Parameters:
        data    -   numpy.ndaray with data to be used as a background data.
    Returns:
        None
"""
        if data is None:
            self.bg = self.bgNone
            return
        if ( len(data) != self.X_SIZE ) :
            self.bg = self.bgNone
            raise RuntimeError("SetBG::Range mismatch")
        try:
            self.bg[:,:] = data.copy()
        except:
            x = np.array(data,dtype=float)
            self.bg[:,:] = np.array(data,dtype=float)


    def SelectBG(self, val):
        """Selects a background data from existing list. The data are set as current backgound data to be used in processing.
    Parameters:
        val    -   index of the data in the list (corresponds to exposure)
    Returns:
        None
"""
        try:
            
            self.SetBG(self.bgList[val].data)
            self.currentBG = val
        except IndexError:
            self.SetBG(None)
            self.currentBG = None

    def GetSelectedBG(self):
        """Gets index of a currently selected background data within the list.
    Parameters:
        None
    Returns:
        index of a currently selected background.
"""
        return self.currentBG

    def SetNorm(self,data=None, mask=None):
        """Sets current normalization data to be used in processing.
    Parameters:
        data    -   numpy.ndaray with data to be used for normalization (default:None).
        mask    -   numpy.ndaray with corresponding mask data to be used for normalization (default:None).
    Returns:
        None
"""
        if data is None:
            self.norm = self.normNone
            self.normSrc = self.normNone
            self.mask = self.maskNone
            self.prepareNorm()
            
            return
        if ( len(data) != self.X_SIZE  ) :
            self.norm = self.normNone
            self.normSrc = self.normNone
            self.mask = self.maskNone
            raise RuntimeError("SetNorm::Range mismatch")
            return
        else:
            self.normSrc[:,:] = data
        if mask is not None:
            try:
                self.mask[:,:] = mask.copy()
            except:
                self.mask[:,:] = np.array(mask,dtype=float)            
        else:
            self.mask = self.maskNone

        self.prepareNorm()

    def GetNormList(self):
        """Gets a list of normalizations as dictionary. It includes all possible keys and boolean values indicate whether the corresponding normalizaion is available at the moment.
    Parameters:
        None
    Returns:
        dictionary {"default":True|False, "recorded":True|False}
"""
        return dict(((key, key in self.normDict) for key in DEFAULT_NORM_KEYS))

    def GetSelectedNorm(self):
        """Gets key for currently selected normalization.
    Parameters:
        None
    Returns:
        key for currently selected normalization ("default" or "recorded") or None if none is selected.
    """
        return self.currentNorm

    def SelectNorm(self,val = None):
        """Selects normalization from the dictionary by the key (either "default", "recorded", or None). 
    Parameters:
        val -   key value. If the key is None or does not exists empty normalization is switched off.
    Returns:
        The key value on success or None otherwise.
    """
        if val in self.normDict:

            self.SetNorm(self.normDict[val].data,self.normDict[val].mask)
            self.currentNorm = val
        else:
            self.SetNorm()
            self.currentNorm = None
        return self.currentNorm
        



    def SetDifference(self, on = True):
        """Turns on difference mode on or off."""
        if on:
            self.interface.diff_mode(True)            
            self.processFlags |= DIFFERENCE
##            self.parity = True
##            self.odd_frame = np.zeros((self.X_SIZE,self.Y_SIZE),dtype=float)

        else:
            self.interface.diff_mode(False)
            self.processFlags &= ~DIFFERENCE
            
        
    def LoadConfig(self,zipName=None,pulseCallback=None):
        """Loads background and normalization data from a configuration file and puts in corresponding list and dictionary.
    Parameters:
        zipName  -   filename for a configuration file. If it is not provided, default configuration file is loaded.
        pulseCallback   -       callback to indicate progress (it is called without parameters).
    Returns:
        None

    If filename is provided as a relative path, it is relative to the module folder.
"""

        bgList=[]
        normDict={}

        configZipName = zipName or "data/"+self.interface.get_id()+".zip"

        with zipfile.ZipFile(configZipName) as configZip:
            if callable(pulseCallback):
                pulseCallback()
            for name in DEFAULT_BG_FILENAMES:                
                with configZip.open(name,"r") as f:
                    file_in=io.BytesIO(b"\n".join(f.read().splitlines()))
                    data = ref.RefData(self.X_SIZE,1, style = ref.BG)
                    data.load(file_in)
                    file_in.close()
                bgList.append(data)
                
                if callable(pulseCallback):
                    pulseCallback()

            for label,name in DEFAULT_NORM_FILENAMES.items():
                if name is None:
                    continue
                with configZip.open(name,"r") as f:
                    file_in=io.BytesIO(b"\n".join(f.read().splitlines()))
                    data = ref.RefData(self.X_SIZE,1, style = ref.NORM)
                    data.load(file_in)
                    file_in.close()
                normDict[label] = data
                
                if callable(pulseCallback):
                    pulseCallback()
                        
        self.bgList = bgList
        self.normDict = normDict
        self.SelectBG(0)
        s = self.GetSelectedNorm()
        if s in self.normDict:
            self.SelectNorm(s)
        else:
            self.SelectNorm('default')
        if callable(pulseCallback):
            pulseCallback()
            
    def LoadConfigFromMemory(self, data):
        """ Decode binary config and load it"""
        assert isinstance(data,bytes) or isinstance(data,bytearray)
        self.MakeNullBGList()
        self.MakeNullNormList()
        header_format = b'6sHBB'
        data_format = b'h'
        if len(data) < struct.calcsize(header_format):
            raise RuntimeError("Too little calibration data.")
        tag, size, nN, bN = struct.unpack(header_format, data[: struct.calcsize(header_format)])
        
        if not ( (tag==b"TS_CAL" or tag==b"TF_CAL") and size == self.X_SIZE
                 and nN == len(self.normDict) and bN==len(self.bgList)
                     and len(data) == struct.calcsize(header_format)+ struct.calcsize(data_format)*size*(2*nN+bN)):
            raise  RuntimeError("Invalid calibration data.")
        
        data_iter = struct.iter_unpack(data_format,data[struct.calcsize(header_format):])

        j=0
        for key in ("default","recorded"):
            value = self.normDict[key]
            for i in range(self.X_SIZE):
                
                value.data[i][j] = next(data_iter)[0]/SCALE_NORM                    
            for i in range(self.X_SIZE):
                value.mask[i][j] = next(data_iter)[0]/SCALE_MASK

        for value in self.bgList:
            for i in range(self.X_SIZE):
                value.data[i][j] = next(data_iter)[0]/SCALE_BG
                 
        


    def SaveConfig(self,filename,pulseCallback=None):
        """Saves complete list of backgrounds and normalization data dictionary to a configuration file.
    Parameters:
        filename  -   filename for a configuration file (required parameter).
        pulseCallback   -       callback to indicate progress (it is called without parameters).
    Returns:
        None

    If filename is provided as a relative path, it is relative to the module folder.
"""
        if callable(pulseCallback):
            pulseCallback()
        with zipfile.ZipFile(filename,"w") as configZip:
            for i,name in enumerate(DEFAULT_BG_FILENAMES):
                
                if i >= len(self.bgList):
                    break
                name = DEFAULT_BG_FILENAMES[i]
                file_out = tempfile.NamedTemporaryFile(delete=False)
                tempfilename = file_out.name
                temp = np.array(self.bgList[i].data)
                data = ref.RefData(self.X_SIZE,1,data=temp[:,0].reshape((self.X_SIZE,1)), style = ref.BG, exposure = 0)
                data.save(file_out)
                file_out.close()
                configZip.write(tempfilename,name,zipfile.ZIP_DEFLATED)
                os.remove(tempfilename)
                if callable(pulseCallback):
                    pulseCallback()

            for  label,name in DEFAULT_NORM_FILENAMES.items():
                if label in self.normDict and name is not None:
                    file_out = tempfile.NamedTemporaryFile(delete=False)
                    tempfilename = file_out.name
                    temp = np.array(self.normDict[label].data)
                    temp1 = np.array(self.normDict[label].mask)
                    data = ref.RefData(self.X_SIZE,1,data=temp[:,0].reshape((self.X_SIZE,1)), mask = temp1[:,0].reshape((self.X_SIZE,1)), style = ref.NORM, exposure = 0)
                    data.save(file_out)
                    file_out.close()
                    configZip.write(tempfilename,name,zipfile.ZIP_DEFLATED)
                    os.remove(tempfilename)
                    if callable(pulseCallback):
                        pulseCallback()

    def RecordCurrentBG(self,count=30, ticker=None):
        """Record background data for current exposure.
    Parameters:
        count  -   number of repetitions used for averaging (default: 30).
    Returns:
        teraFAST.ref.RefData object with the background information.
"""
        data = np.zeros((self.X_SIZE,self.Y_SIZE),dtype=float)
        res = np.empty((self.X_SIZE,1),dtype=float)
##        dur = self.GetIntTime()
        dur = 1
        i=0
        while i <count:
            tmp = self.readBGData()
            if tmp is None:
                continue
            i+=1
            np.add(data,tmp,data)
            
            if callable(ticker):
                cont, skip = ticker(dur)
                if not cont:
                    return None
        np.divide(data,count,data)
        tmp = np.sum(data, axis=1,keepdims=True)
        res[:,:]=np.divide(tmp,self.Y_SIZE)
        return ref.RefData(self.X_SIZE,1,data=res, style = ref.BG, exposure = 0)
   
    def RecordBG(self,count=0,callback=None):
        """Record background data for all available exposures.
    Parameters:
        count  -   number of repetitions used for averaging. If 0 or not provided, then default value is used (30, may be changed in defaults file).
        callback    -   callback to indicate progress. It is called with completed percentage and it is expected to return tuple (continue, skip), where continue is True unless the process should be aborted (see wx.ProgressDialog)
    Returns:
        True if completed successfully, False if canceled
"""
        flag = False
        cancelled = False
        if self.proc_thread is None or not self.proc_thread.is_alive():
            flag = True
            self.start()
            self.read()        
                


        if not count:
            try:
                count = self.recordingCount
            except AttributeError:
                count = 30

        pr = progress(count,callback)
        result = []

        x = self.RecordCurrentBG(count, pr.increase)
        if x is not None:
            result.append(x)
        else:
            cancelled=True
##            break
            
        if not cancelled:   
            self.bgList = result
##        self.SetExposure(current_exp)
        self.SelectBG(0)
        if flag:
            self.stop()

        return (not cancelled)



    
    def RecordNorm(self,count=0,callback=None, force=False):
        """Record normalization data.
    Parameters:
        count  -   number of repetitions used for averaging. If 0 or not provided, then default value is used (30, may be changed in defaults file).
        callback    -   callback to indicate progress. It is called with completed percentage and it is expected to return tuple (continur, skip), where continue is True unless the process should be aborted (see wx.ProgressDialog)
    Returns:
        True if completed successfully, False if canceled.
"""
        cancelled = flag = False
        if self.proc_thread is None or not self.proc_thread.is_alive():
            flag = True
            self.start()
            self.read()

        
        if not count:
            try:
                count = self.recordingCount
            except AttributeError:
                count = 30

        
        data = np.zeros((self.X_SIZE,self.Y_SIZE),dtype=float)
        squared = np.zeros((self.X_SIZE,self.Y_SIZE),dtype=float)
        res = np.empty((self.X_SIZE,1),dtype=float)
        sd2 = np.empty((self.X_SIZE,1),dtype=float)
        pr = progress(count,callback)

         
        i=0
        while i <count:
            tmp = self.readNormData()
            if tmp is None:
                continue
            i+=1
            np.add(data,tmp,data)
            np.square(tmp,tmp)
            np.add(squared,tmp,squared)
            
            cont, skip = pr.increase(1)

            if not cont:
                cancelled = True
                break
        if not cancelled:   
            np.divide(data,count,data)
            tmp = np.sum(data, axis=1,keepdims=True)
            res[:,:]=np.divide(tmp,self.Y_SIZE)
            
            tmp = np.sum(np.divide(squared,count), axis=1,keepdims=True)
            sd2[:,:] = np.subtract(np.divide(tmp, self.Y_SIZE),np.square(res))

            if np.any(sd2 < 0):
                a = np.extract(sd2<0,sd2)
                raise RuntimeError
            test = (sd2 == 0)
            sd = np.sqrt(sd2)
            mask = np.where(test,0, np.divide(np.absolute(res),sd))
            mask = np.where(np.logical_and(test, res!=0), MAX_FLOAT,   mask)

            res /= np.abs(res).max()/float(MAX_VALUE)
            x = ref.RefData(self.X_SIZE, 1, data=res, mask=mask[:,0:1].astype(float), style = ref.NORM, exposure = 0)


            self.normDict['recorded'] = x
            

            self.SelectNorm('recorded')

                  
        if flag:
            self.stop()
       
        return (not cancelled)


    def MakeNullBGList(self):
        
        self.bgList = []
        self.bgList.append(ref.RefData(self.X_SIZE,1,data=self.bgNone[:,0:1], style = ref.BG, exposure = 0))
            


    def MakeNullNormList(self):
        self.normDict ={}
        for  label,name in DEFAULT_NORM_FILENAMES.items():
            self.normDict[label] = ref.RefData(self.X_SIZE,1,data=self.normNone[:,0:1].astype(float), mask=self.maskNone[:,0:1].astype(float), style = ref.NORM, exposure = 0)
    

    
