import numpy as np
import  cv2




class Worker(object):
    """Image generation class. Uses single 2D array of numpy.float [0,1] to produce three-channel numpy.uint8 image.  """
# processFlags
##    FALSECOLOR = 0x1
    SMOOTH     =   0x10
    NEGATIVE   =   0x20
    MEDIAN      = 0x40
    MIRROR  =   0x100
    DRAW_FRAME =   0x400
    DRAW_GRID =   0x800
    WATERFALL =     0x1000
    DRAW_ROI   = 0x8000
    DEFAULT_FLAGS = SMOOTH

    WF_ADD = 0x1
    WF_ADD_AVERAGE = 0x2
    WF_COMPRESS = 0x4
    
    def __init__(self, size=(32,32), flags = None):
        """Initialize object.
    Parameters:
        size    -   dimensions of the image width, height (default: (32,32))
        flags   -   processing flags
        wfSize  -   dimensions of the waterfall image mode. If None, square image is produced

    Processing flags:
        FALSECOLOR  -   produce image in false colors (rainbow) instead of b/w (tinted)
        SMOOTH     -    smooth image (space-domain filtering)
        NEGATIVE   -    invert image
        MEDIAN      -   use median filtering instead of gaussian blurring
        MIRROR  -   mirror image
        DEFAULT_FLAGS = FALSECOLOR | SMOOTH
"""
        	
        self.processFlags = flags or self.DEFAULT_FLAGS

        self.wfModesDict = {"Add":self.WF_ADD,"Add average":self.WF_ADD_AVERAGE}
        
        self.wfMode = self.WF_ADD
        self.wfFlags = 0x0


        self.max, self.average, self.bins = 0,0,[]
        
        self.brightness = 0 # -1, +1
        self.contrast = 1  # 0-inf
        self.SetGamma(1)   # 0-2
        
        self.smoothness = 0.5

        self.compress = 1
        self.resizeRatio = 1
        self.size = size
        self.dimX, self.dimY = size

        self.img = np.empty( self.size+(3,),dtype=np.uint8)
        self.source = np.zeros( self.size, dtype=float)
        self.im_src = np.zeros( self.size, dtype=float)
        self.ntc = 0
        self.dtc = 0

        self.lut = self.generateLUT()
        

            
    def makeImg(self, data):
        """Generates image according to current settings.
    Parameters:
        data -   input data
    Returns:
        image
"""
        
        self.data = data.transpose().copy()
        if self.resizeRatio != 1:
            if self.resizeRatio * self.data.shape[0] >= 1:
                self.data = cv2.resize(self.data,(0,0),fx=1,fy=self.resizeRatio,interpolation=cv2.INTER_AREA)
            else:
                self.data = cv2.resize(self.data,(self.data.shape[1],1),interpolation=cv2.INTER_AREA)
        if self.source.shape[0] >= self.data.shape[0]:
            self.source[self.data.shape[0]:,:] =  self.source[0:-self.data.shape[0],:]
            self.source[0:self.data.shape[0],:] = self.data
        else:
            self.data = cv2.resize(self.data,(self.source.shape[1],self.source.shape[0]),interpolation=cv2.INTER_AREA)
            self.source[:,:] = self.data
            
        self.data   =  self.generateImage(self.data)

        self.im_src[self.data.shape[0]:,:] =  self.im_src[0:-self.data.shape[0],:]
        self.im_src[0:self.data.shape[0],:] = self.data  

        if self.wfMode == self.WF_ADD or self.wfMode == self.WF_COMPRESS:
            sample = self.data2RGB(self.data)            
        elif self.wfMode == self.WF_ADD_AVERAGE:
            self.data[0,:] = self.data.mean(0)
            sample = self.data2RGB(self.data[0:1,:])[0:1,:,:]
        self.img[sample.shape[0]:,:] =  self.img[0:-sample.shape[0],:]
        self.img[0:sample.shape[0],:] = sample                
            

        return self.img

    
##    def SetWaterfall(self, on = True, wfRatio = None):
##        """Turns waterfall mode on or off.
##    Parameters:
##        on   -  switch to turn the waterfall mode either on or off. Default = True
##    Returns:
##        (width, height) - shape of the image after the action
##
##"""
##        if on:
##            self.processFlags |= self.WATERFALL
##            if wfRatio is not None:
##                self.wfSize = (int(self.size[1]*wfRatio), self.size[1])
##            self.img = np.zeros( self.wfSize+(3,),dtype=np.uint8)
##            
##        else:
##            self.processFlags &= ~self.WATERFALL
##            self.img = np.zeros( self.size+(3,),dtype=np.uint8)
##            self.wfSize = (int(self.size[1]/2), self.size[1])
##
##        return self.img.shape[0:2]

    def SetResizeRatio(self,x):
        self.resizeRatio = x if x > 0 else 1
    
    def SetBrightness(self,black,white):
        """Sets brightness using black and white points.
    Parameters:
        black   -   black point value in [0, white)
        white   -   white point value in (black, 1]
    Returns:
        None
"""
        self.brightness = float(black)

    def SetContrast(self,black,white):
        """Sets contrast using black and white points.
    Parameters:
        black   -   black point value
        white   -   white point value 
    Returns:
        None
"""
        self.contrast =  1/float(white-black) if white-black>0.01 else 100


    def SetGamma(self,val):
        """Sets gamma value.
    Parameters:
        val   -   gamma value (gamma > 0)        
    Returns:
        None
"""
        self.gamma = val if val > 0 else 0
        self.gamma_lut = np.clip(np.power(np.arange(256,dtype=float)/255,self.gamma)*255,0,255).astype(np.uint8)

    def SetSmoothness(self, val):
        """Sets smoothness value.
    Parameters:
        val   -   smoothness value (0, 100]        
    Returns:
        None
"""
        self.smoothness=float(val)/100

    def GetSmoothness(self):
        """Gets smoothness value.
    Parameters:
        None   
    Returns:
        smoothness value (0, 100]
"""
        return int(self.smoothness*100)

    def generateLUT(self,mode="fc"):
        """ Generates look-up table to convert a value in 0-255 range into color.
        Parameters:
            mode --- either "fc" (false-color), or "bw" (blue-tinted black&white).
        Returns a tuple with three lookup tables for R, G, and B channels."""

        xxx = np.arange(256,dtype=np.uint8)

        temp = np.zeros((256,1,3),dtype=np.uint8)

        if mode == "bw":
            temp[:,0,0]=120
            temp[:,0,1]=xxx
            temp[:,0,2]=255
            cv2.cvtColor(temp,cv2.COLOR_HLS2RGB, temp)


        elif mode == "fc":
            temp[0:64,0,0]= 0
            temp[0:64,0,1]=0
            temp[0:64,0,2]= (xxx[0:64]*255./64).astype(np.uint8)

            temp[64:96,0,0]= 0
            temp[64:96,0,1]= (xxx[0:32]*255./32).astype(np.uint8)
            temp[64:96,0,2]= 255

            temp[96:128,0,0]= 0
            temp[96:128,0,1]= 255
            temp[96:128,0,2]= (255 - xxx[0:32]*255./32).astype(np.uint8)

            temp[128:192,0,0]= (xxx[0:64]*255./64).astype(np.uint8)
            temp[128:192,0,1]= 255
            temp[128:192,0,2]= 0

            temp[192:256,0,0]= 255
            temp[192:256,0,1]= (255 - xxx[0:64]*255./63).astype(np.uint8)   # To get (255, 0, 0) and i=255
            temp[192:256,0,2]= 0


        else:
            temp[:,0,0]=xxx
            temp[:,0,1]=xxx
            temp[:,0,2]=xxx
            
        return (temp[:,0,0].copy(),temp[:,0,1].copy(),temp[:,0,2].copy())

    

    def SetFC(self,mode=True):
        if mode:
            self.lut = self.generateLUT("fc")
        else:
            self.lut = self.generateLUT("bw")
        
    def generateImage(self,data):
        """Applies image-processing operations to the original data. Data processing is performed in-place.
    Parameters:
        data    -   input data
    Returns:
        output data (in most cases it is the same array).
"""
 
        if (self.processFlags & self.MIRROR):
            data=np.fliplr(data)

        x = np.clip((255*self.contrast*(data-self.brightness)),0,255).astype(np.uint8)
        

        if self.gamma != 1:
            cv2.LUT(x,self.gamma_lut,x)
        if (self.processFlags & self.NEGATIVE):
            np.subtract(255, x, x)
           
        
        if (self.processFlags & self.SMOOTH and self.smoothness > 0) :
            if (self.processFlags & self.MEDIAN):
                k = 3 if self.smoothness <=0.5 else 5 
                cv2.medianBlur(x,k,x)
            else:            
                cv2.GaussianBlur(x,(3,3),self.smoothness,x)
                
        
        
        return x

    
    def data2RGB(self,data,contours=None):
        """Converts one-channel BW data to three-channel RGB data, output depends on FALSECOLOR flag.
    Parameters:
        data    -   input data
    Returns:
        A link to self.img, which contains resulting image
    
"""

        temp = np.empty(data.shape+(3,),dtype=np.uint8)        
        temp[:,:,0] = cv2.LUT(data,self.lut[0])
        temp[:,:,1] = cv2.LUT(data,self.lut[1])
        temp[:,:,2] = cv2.LUT(data,self.lut[2])
        return temp




    def statistics(self, selection=(None,None)):
        """Calculates simple statistics on data or a region of the data (mean, minimum, and maximum values).
    Parameters:
        data    -   input data
        selection   -   ((x1,y1), (x2,y2)) coordinates of two corners of the region to be used. If omitted, entire array is used/
    Returns:
        A dictionary {"mean":float, "min":float, "max":float}
"""

        a, b = selection
        if isinstance(a,tuple) and isinstance(b,tuple):
            x = self.source[a[0]:b[0]+1,a[1]:b[1]+1]
        else:
            x = self.source
        mean = np.mean(x)
        amin = round(np.amin(x),3)
        amax = round(np.amax(x),3)
            
        return {"mean":mean, "min":amin, "max":amax}

        
